# By transforming the colour image in CIE Lab* colour space, calculating the chromaticity and luminance differences of adjacent pixels.
# Then applying a non-linear function to obtain the greyscale target differences, and adjusting the greyscale map by solving an optimisation problem.
# Thus the goal of converting colour images to greyscale images and preserving the perceived differences between colours is achieved.
import os
import sys
import argparse
from numpy import linalg as LA
import numpy as np
import scipy
from scipy import io, sparse
from skimage import color
import imageio
#import imageio.v2 as imageio
from tqdm import tqdm

alpha = 8  # The parameter alpha is used to calculate the colour difference
theta = np.pi / 4  # The parameter theta is used to calculate the colour difference
v_theta = np.asarray((np.cos(theta), np.sin(theta)))  # The parameter v_theta is used to calculate the colour difference

# Compute the dot product of two arrays
def dot_prod(arrays, out=None):
    arrays = [np.asarray(x) for x in arrays]
    dtype = arrays[0].dtype
    n = np.prod([x.size for x in arrays])
    if out is None:
        out = np.zeros([n, len(arrays)], dtype=dtype)
    m = n // arrays[0].size
    out[:, 0] = np.repeat(arrays[0], m)
    if arrays[1:]:
        dot_prod(arrays[1:], out=out[0:m, 1:])
        for j in range(1, arrays[0].size):
            out[j * m:(j + 1) * m, 1:] = out[0:m, 1:]
    return out

# Functions: calculating the colour difference
# In the crunch function, the colour difference is handled by applying the non-linear
# function crunch(x) = alpha * tanh(x / alpha), where the parameter alpha controls the intensity of the effect.
# In this way, the final greyscale value of each pixel is obtained.
def crunch(x):
    return alpha * np.tanh(x / alpha)


# Find the neighbouring pixels of a pixel
def neighbour_pixels(pos_mat, r, mu=1):
    d = mu
    l1, h1 = max(r[0] - d, 0), min(r[0] + d, pos_mat.shape[0])
    l2, h2 = max(r[1] - d, 0), min(r[1] + d, pos_mat.shape[1])
    return pos_mat[l1:h1 + 1, l2:h2 + 1]


# Calculate the difference between a pixel and its neighbouring pixels in LAB colour space
def delta_calculation(col_lab_result, r, S):
    values = []
    for s in S:
        delta_luminance = col_lab_result[r][0] - col_lab_result[s][0]  # Calculate the luminance difference
        delta_chrominance = col_lab_result[r][1:] - col_lab_result[s][1:]  # Calculation of chromaticity difference
        delta_cnorm_res = LA.norm(delta_chrominance)  # Calculating the Paradigm of Chromaticity Difference
        cval = crunch(delta_cnorm_res)  # Calculate the colour difference
        dotprod = np.dot(v_theta, delta_chrominance)  # Calculate the dot product of the chromaticity difference and the parameter v_theta
        if abs(delta_luminance) > cval:
            values.append(delta_luminance)
        elif dotprod >= 0:
            values.append(cval)
        else:
            values.append(crunch(-delta_cnorm_res))
    return values


# Optimisation calculations
# In the do_optimization function, the greyscale map is adjusted by solving an optimization problem.
# where matrices A and B are used to construct the coefficient matrix of the optimisation problem and the optimisation problem is solved by np.linalg.solve. The final best_g obtained is the adjusted grey map.
def do_optimization(delta, size):
    A = np.ones((size, size), dtype='float32') * -1.0
    for i in range(size):
        A[i, i] = size
    B = np.sum(delta, axis=0).transpose() - np.sum(delta, axis=1)
    return A, B


#Calculate the colour difference of all pixels
def create_delta(col_lab_result, list_neighbour, arrxy2idx, height, width):
    size = height * width
    delta = sparse.lil_matrix((size, size))
    for i in range(height):
        for j in range(width):
            if j == 0:
                print(f"Processing row {i+1} of {height}")
            cur_index = arrxy2idx[i, j]
            neighbors = list_neighbour[cur_index]
            n_indices = [arrxy2idx[pos] for pos in neighbors]
            cur_val = delta_calculation(col_lab_result, (i, j), neighbors)
            delta[cur_index, n_indices] = np.asmatrix(cur_val)
    return delta.todense()

def main(argts):
    par = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    par.add_argument('--input', help='Specify the input picture.', type=str, required=True)
    par.add_argument('--output', help='Specify the output picture.', type=str, required=True)

    args = par.parse_args(argts)

    input_path = args.input
    output_path = args.output

    col_rgb = imageio.imread(input_path)
    # Use the colour.rgb2lab function to convert the RGB input image col_rgb to col_lab_result in the CIE Lab* colour space so that the converted image better represents the perceived differences between colours.
    col_lab_result = color.rgb2lab(col_rgb[:, :, 0:3])

    (height, width) = col_rgb.shape[0:2]
    size = height * width

    gray_matrix = np.average(col_rgb, axis=2)

    cart = dot_prod([range(height), range(width)])
    cart_r = cart.reshape(height, width, 2)
    arrxy2idx = np.arange(size).reshape(height, width)
    list_neighbour = []  # Store the neighbouring pixels of each pixel
    for i in tqdm(range(height), desc="Processing rows"):
        for j in range(width):
            cur_index = arrxy2idx[i, j]
            cur_neighbor = neighbour_pixels(cart_r, [i, j], max(height, width)).reshape(-1, 2)  #Considering both near and far pixels, the neighbourhood of each pixel is essentially the entire image
            cur_neighbor = [tuple(item) for item in cur_neighbor]
            cur_neighbor.remove((i, j))
            list_neighbour.append(cur_neighbor)
            # In the create_delta function, a sparse matrix delta is first defined to store the difference in grey targets between pixels. Then by traversing each pixel, the
            # Calculate the chromaticity difference and luminance difference between it and its neighbouring pixels, and determine the greyscale target difference based on these differences.
            # where the delta_calculation function is used to calculate the chromaticity difference and luminance difference between neighbouring pixels, and the non-linear function crunch(x) is applied to determine the greyscale target difference
    delta_matrix = create_delta(col_lab_result, list_neighbour, arrxy2idx, height, width) # Calculate the colour difference of all pixels

    A, B = do_optimization(delta_matrix, size)
    best_g = np.linalg.solve(A, -B)
    g_matrix = best_g.reshape(height, width)
    gray_matrix = np.zeros((height, width, 3))
    for i in range(3):
        gray_matrix[:, :, i] = g_matrix

    imageio.imsave(output_path, gray_matrix)
    print('Color picture saved to', output_path)

if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))


#Run command
# python test.py --input "D:/color_gray/rgb_gray/rgb_gray/image3/1.png" --output "D:/color_gray/rgb_gray/rgb_gray/image3/1_label.png"

# python test.py --input "D:/color_gray/rgb_gray/rgb_gray/image3/001.png" --output "D:/color_gray/rgb_gray/rgb_gray/image3/001_label.png"